# -*- coding: utf8 -*-

#-----------------------------------------
# This is a part of Collège 3D project.
# Author:       Pascal Peter
# Copyright:    (C) 2003-2016 Pascal Peter
# Licence:      GNU General Public Licence version 3
# Website:      http://pascal.peter.free.fr/
# Email:        pascal.peter at free.fr
#-----------------------------------------

"""
DESCRIPTION :
    Ce module contient les variables et fonctions utiles au programme.
"""

# importation des modules utiles :
import sys
import os
import random
import csv

import bge
from bgl import *
import blf



###########################################################'
#
#       CONSTANTES DIVERSES :
#
###########################################################'

PROGNAME = 'Collège 3D'
PROGYEAR = '2015-2016'
PROGAUTHOR = 'Pascal Peter'
PROGMAIL = 'pascal.peter at free.fr'
PROGWEB = 'http://clgdrouyn.fr/c3d'
LICENCETITLE = 'GNU General Public License version 3'
SHORT_NAME = 'college3d'
WEBPAGE_HELP = 'http://clgdrouyn.fr/c3d/help.html'
WEBPAGE_ABOUT = 'http://clgdrouyn.fr/c3d'
DOWNLOAD_DIR = 'http://clgdrouyn.fr/c3d/wikiuploads'

SEP = os.sep

LAN_PORT = 5005

try:
    BGE_VERSION = bge.app.version[0] * 100 + bge.app.version[1]
except:
    BGE_VERSION = 273

###########################################################'
#
#       CONFIGURATION DU JEU :
#
###########################################################'

defaultSettingsValues = {
    'MENU': {
        'gameName': '', 
        'character': 'eleve', 
        'mode': 'keyboard', 
        'lang': 'fr', 

        'fullScreen': False, 
        'anaglyphs': False, 
        'sounds': True, 
        'music': True, 
        'infos': False, 
        'video': False, 
        },

    'LAN': {
        'host': 'localhost', 
        'user': '', 
        'password': '', 
        },

    'OTHER': {
        'lastDownload': 10000101, 
        'installedPython': '', 
        'bundledPython': '', 
        },

    'GAME': {},

    'TEMP': {},
    }

def findConfigDir():
    """
    retourne l'emplacement du fichier de configuration
    """
    configDir = ''
    if sys.platform.startswith('win'):
        configDir = os.environ['APPDATA']
        configDir = os.path.join(configDir, SHORT_NAME)
    else:
        configsDir = os.path.join(os.path.expanduser('~'), '.config')
        if os.path.isdir(configsDir):
            configDir = os.path.join(configsDir, SHORT_NAME)
        else:
            configDir = os.path.join(os.path.expanduser('~'), '.{0}'.format(SHORT_NAME))
    # on vérifie l'existence du dossier :
    if not(os.path.isdir(configDir)):
        os.mkdir(configDir)
    if not(os.path.isdir(configDir)):
        configDir = bge.logic.expandPath('//')
    #print('configDir :', configDir)
    return configDir

def loadSettings():
    """
    chargement de la configuration
    """
    #print(bge.logic.globalDict)
    # pour n'appeler cette fonction qu'une seule fois :
    if 'configFile' in bge.logic.globalDict:
        return
    import datetime
    import copy
    import scenarios
    # on commence par lire le fichier de config :
    configDir = findConfigDir()
    configFile = os.path.join(configDir, 'college3d.json')
    configDict = {}
    try:
        import json
        if os.path.isfile(configFile):
            theFile = open(configFile, newline='', encoding='utf-8')
            configDict = json.load(theFile)
            theFile.close()
    except:
        configDict = {}
    # récupération ou initialisation des valeurs :
    for key in defaultSettingsValues:
        bge.logic.globalDict[key] = defaultSettingsValues[key]
        for key2 in bge.logic.globalDict[key]:
            if key in configDict:
                configDictValue = configDict[key].get(key2, None)
                if configDictValue != None:
                    bge.logic.globalDict[key][key2] = configDictValue
    # on met configFile dans le dico :
    bge.logic.globalDict['configFile'] = configFile
    # on récupère la liste des jeux :
    bge.logic.globalDict['GAMES'] = {'LIST': [], 'INITIAL_POSITIONS': {}, 'WON_SCENES': {}}
    for game in scenarios.games:
        bge.logic.globalDict['GAMES']['LIST'].append(game)
        bge.logic.globalDict['GAMES'][game] = []
        if game in scenarios.levels:
            for level in scenarios.levels[game]:
                bge.logic.globalDict['GAMES'][game].append([copy.deepcopy(level), False])
        if game in scenarios.initialPositions:
            bge.logic.globalDict['GAMES']['INITIAL_POSITIONS'][game] = copy.deepcopy(
                scenarios.initialPositions[game])
        if game in scenarios.wonScenes:
            bge.logic.globalDict['GAMES']['WON_SCENES'][game] = copy.deepcopy(
                scenarios.wonScenes[game])
    # vérification de certaines valeurs (au cas où elles n'existeraient plus) :
    if not(bge.logic.globalDict['MENU']['gameName'] in  bge.logic.globalDict['GAMES']['LIST']):
        bge.logic.globalDict['MENU']['gameName'] = bge.logic.globalDict['GAMES']['LIST'][0]
    if not(bge.logic.globalDict['MENU']['character'] in  characters):
        bge.logic.globalDict['MENU']['character'] = characters[0]
    if not(bge.logic.globalDict['MENU']['mode'] in  modes):
        bge.logic.globalDict['MENU']['mode'] = modes[0]
    bge.logic.globalDict['TRANSLATIONS'] = {'csvFiles':[], 'languages':[], 'tr':{}}
    openCsvFile('menu')
    languages = bge.logic.globalDict['TRANSLATIONS']['languages']
    if not(bge.logic.globalDict['MENU']['lang'] in  languages):
        bge.logic.globalDict['MENU']['lang'] = languages[0]
    # on efface les niveaux
    bge.logic.globalDict['GAME']['level'] = ''
    # le jeu n'est pas en pause au lancement :
    bge.logic.globalDict['GAME']['pause'] = False
    # et le nom d'utilisateur réseau est vide :
    bge.logic.globalDict['LAN']['user'] = ''
    # screenshots et vidéos :
    try:
        bge.logic.globalDict['MENU']['video'] = ('VIDEO' in sys.argv)
    except:
        bge.logic.globalDict['MENU']['video'] = False
    videoDir = os.path.join(configDir, 'video')
    if not(os.path.isdir(videoDir)):
        os.mkdir(videoDir)
    if not(os.path.isdir(videoDir)):
        videoDir = ''
    bge.logic.globalDict['VIDEO'] = {}
    bge.logic.globalDict['VIDEO']['videoDir'] = videoDir
    bge.logic.globalDict['VIDEO']['NumScreenshot'] = 0
    bge.logic.globalDict['VIDEO']['CompteurScreenshot'] = 0
    if bge.logic.globalDict['MENU']['video']:
        openDir(bge.logic.globalDict['VIDEO']['videoDir'])
    # images téléchargées (jeu Exhibition) :
    downloadDir = os.path.join(configDir, 'download')
    if not(os.path.isdir(downloadDir)):
        os.mkdir(downloadDir)
    if not(os.path.isdir(downloadDir)):
        downloadDir = ''
    bge.logic.globalDict['TEXTURES'] = {}
    bge.logic.globalDict['downloadDir'] = downloadDir
    # position du curseur :
    bge.logic.globalDict['GAME']['cursorPosition'] = (0.5, 0.8)
    # récupération de la date (pour jouer avec les saisons) :
    today = datetime.date.today()
    winter = datetime.date(today.year, 12, 21)
    autumn = datetime.date(today.year, 9, 21)
    summer = datetime.date(today.year, 6, 21)
    spring = datetime.date(today.year, 3, 21)
    if today > winter:
        season = 'Winter'
    elif today > autumn:
        season = 'Autumn'
    elif today > summer:
        season = 'Summer'
    elif today > spring:
        season = 'Spring'
    else:
        season = 'Winter'
    calendar = calendarObject(today.year, today.month, today.day)
    bge.logic.globalDict['SEASONS'] = {
        'today': (today.year, today.month, today.day),
        'calendar': calendar,
        'season': season}
    (bundledPython, installedPython) = detectPython()
    bge.logic.globalDict['OTHER']['bundledPython'] = bundledPython
    bge.logic.globalDict['OTHER']['installedPython'] = installedPython

def detectPython():
    """
    recherche le chemin des executables Python :
        * celui embarqué avec le jeu
        * celui installé sur l'ordinateur.
    """
    # Python embarqué :
    bundledPython = ''
    if sys.platform.startswith('linux'):
        executableFilename = 'python3.5m' # à améliorer (détecter)
        pythonVersionDir = '2.77' # à améliorer (détecter)
        fileName = '//..{0}..{0}{1}{0}python{0}bin{0}{2}'.format(
            SEP, pythonVersionDir, executableFilename)
        fileName = bge.logic.expandPath(fileName)
        if os.path.exists(fileName):
            bundledPython = fileName
    elif sys.platform.startswith('win'):
        executableFilename = 'python.exe'
        pythonVersionDir = '2.77' # à améliorer (détecter)
        fileName = '//..{0}..{0}{1}{0}python{0}bin{0}{2}'.format(
            SEP, pythonVersionDir, executableFilename)
        fileName = bge.logic.expandPath(fileName)
        if os.path.exists(fileName):
            bundledPython = fileName

    # Python installé sur l'ordinateur :
    installedPython = ''
    if sys.platform.startswith('win'):
        try:
            import winreg
            # on cherche où est python via la base de registre :
            key = winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE,
                'SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\Python.exe',
                0, winreg.KEY_READ)
            (valeur, typevaleur) = winreg.QueryValueEx(key, '')
            winreg.CloseKey(key)
            installedPython = valeur
            # les windowsiens ont peur de la console :
            installedPython = installedPython.replace('ython.exe', 'ythonw.exe')
        except:
            installedPython = ''
    else:
        installedPython = 'python'
    print('bundledPython :', bundledPython)
    print('installedPython :', installedPython)
    return (bundledPython, installedPython)

def saveSettings():
    """
    sauvegarde de la configuration.
    On ne sauvegarde que ce qui est dans defaultSettingsValues.
    """
    import json
    configFile = bge.logic.globalDict['configFile']
    theFile = open(configFile, 'w', encoding='utf-8')
    globalDict = {}
    for key in defaultSettingsValues:
        globalDict[key] = {}
        if not(key in ('GAME', 'TEMP')):
            for key2 in defaultSettingsValues[key]:
                value = bge.logic.globalDict[key].get(key2, defaultSettingsValues[key][key2])
                globalDict[key][key2] = value
    json.dump(globalDict, theFile, indent=4)
    theFile.close()

def clearGlobalDict():
    """
    pour nettoyer le dictionnaire des clés devenues inutiles
    """
    bge.logic.globalDict['TEMP'] = {}



###########################################################'
#
#       PERSONNAGES ET AUTRES TRUCS :
#
###########################################################'

# liste des personnages :
characters = [
    'eleve', 
    'prof', 
    'chien', 
    #'tour', 
    #'test', 
    #'robot', 
    #'kangourou', 
    'moupton', 
    'pingus', 
    'Nono_Baker', 
    ]

# liste des actions :
KX_ACTION_MODE_LOOP = bge.logic.KX_ACTION_MODE_LOOP
actions = {
    'wait': ('wait', 1, 100, KX_ACTION_MODE_LOOP), 
    'walk': ('walk', 1, 20, KX_ACTION_MODE_LOOP), 
    'jump': ('jump', 1, 60, KX_ACTION_MODE_LOOP), 
    }

# liste des modes (clavier, souris, manette, ...) :
if bge.logic.joysticks[0] != None:
    modes = ('keyboard', 'mouse', 'joystick')
else:
    modes = ('keyboard', 'mouse')

# la météo du jeu :
seasonsList = ('Spring', 'Summer', 'Autumn', 'Winter')
weatherDict = {
    'LIST' : ('good weather', 'fog', 'rain', 'snow', 'dead leaves', 'night', 'fish', ),
    'good weather': ('', 0, 0), 
    'fog': ('brouillard', 10, 40), 
    'rain': ('pluie', 100, 200), 
    'snow': ('flocon', 50, 100), 
    'dead leaves': ('feuille', 50, 100), 
    'night': ('nuit', 10, 40), 
    'fish': ('poisscaille', 30, 100), 
    }

def chooseWeather():
    """
    choix d'une météo en fonction de la saison
    """
    season = bge.logic.globalDict['SEASONS']['season']
    if season == 'Spring':
        whatList = ('good weather', 'good weather', 'fog', 'rain', 'rain', 'night',)
    elif season == 'Summer':
        whatList = ('good weather', 'good weather', 'good weather', 'rain', 'night',)
    elif season == 'Autumn':
        whatList = ('good weather', 'fog', 'rain', 'rain', 'dead leaves', 'dead leaves', 'night',)
    elif season == 'Winter':
        whatList = ('fog', 'fog', 'rain', 'snow', 'snow', 'night',)
    weather = weatherDict[random.choice(whatList)]
    objetName = weather[0]
    total = random.randint(weather[1], weather[2])
    return (objetName, total)

def calendarObject(year, month, day):
    """
    pour charger un élément du décor en fonction de la date.
    Par exemple un sapin de Noël.
    Possible aussi de ne mettre des personnages qu'à certaines périodes
    (kangourou en mars, au moment du concours).
    """
    objet = None
    if (month == 12) or ((month == 1) and (day < 20)):
        # Noël (un sapin à l'entrée du collège) :
        objet = ('objets/saisons', 'sapin', (14, 28, 8))
    elif (month == 4) and (day < 11):
        # poisson d'avril :
        objet = ('objets/saisons', 'fish', (14, 28, 8))
    if (month == 3):
        characters.append('kangourou')
    return objet

# liste des jeux en réseau :
lanGames = ('nothing', 'wolf', 'zombies', 'race', 'treasure')
# positions possibles de la cible de la course :
ciblePositions = ((13, 47, 12), (12, 12, 10), (16, -9, 10),)
# positions possibles du trésor :
tresorPositions = ((13, 47, 12), (12, 12, 10), (16, -9, 10),)



###########################################################'
#
#       LES ÉLÉMENTS DU DÉCOR :
#
###########################################################'

# toujours affichés (chargés au lancement) :
allwaysObjects = {
    'ALL_GAMES': (
        'Terrain', 
        'Velos', 
        'Tour', 
        'Hall', 
        'Passerelles', 
        'Bat_A/Batiment_A', 
        'Bat_B/Batiment_B', 
        'Bat_C/Batiment_C', 
        'Bat_D/Batiment_D', 
        ), 
    'Exhibition': ('expo/Expo',), 
    }

# les détails (affichés si on est assez près, d'où les coordonnées et la distance d'affichage) :
detailsObjects = (
    ('Velos_details', (  7,  29,  9), 30), 
    ('Tour_details', ( 16, -25,  9), 30), 
    ('Hall_details', ( 11,   8,  8), 25), 
    ('PasserelleAB_details', (-30,  -8,  8), 20), 
    ('PasserelleAC_details', (  4, -22,  9), 20), 
    ('Bat_A/A_casiers1', (-13,   6,  8), 20), 
    ('Bat_A/A_int_0', (-20,  12,  8), 40), 
    ('Bat_A/A_int_1', (-20,  12, 10), 40), 
    ('Bat_B/B_int_0', (-40, -10,  8), 40), 
    ('Bat_B/B_int_1', (-40, -10, 10), 30), 
    ('Bat_C/C_int_0', (-10,   8,  8), 40), 
    ('Bat_C/C_int_1', (-17, -24, 11), 30), 
    ('Bat_D/D_int_0', ( 25,   0,  9), 40), 
    )



###########################################################'
#
#       POUR CHARGER OU LIBÉRER LES DÉCORS :
#       d'après :
#       http://blenderartists.org/forum/showthread.php?205346-Simple-test-files-for-Dynamic-Loading-
#
###########################################################'

def loadLib(libName, mustReload=False):
    """
    ouvre un fichier placé dans le dossier data
    """
    if mustReload:
        freeLib(libName)
    root = bge.logic.expandPath('//')
    libFile = SEP.join(libName.split('/'))
    libFile = '{1}..{0}data{0}{2}.blend'.format(SEP, root, libFile)
    libFile = bge.logic.expandPath(libFile)
    if not(libFile in bge.logic.LibList()):
        try:
            if libName in detailsObjects:
                bge.logic.LibLoad(libFile, 'Scene', async=True)
            else:
                bge.logic.LibLoad(libFile, 'Scene')
            print('Loaded : {0}'.format(libName))
        except:
            print('ERROR : {0}'.format(libName))
    else:
        print('Library {0} already loaded'.format(libName))

def freeLib(libName):
    """
    libère un fichier
    """
    root = bge.logic.expandPath('//')
    libFile = SEP.join(libName.split('/'))
    libFile = '{1}..{0}data{0}{2}.blend'.format(SEP, root, libFile)
    libFile = bge.logic.expandPath(libFile)
    if libFile in bge.logic.LibList():
        bge.logic.LibFree(libFile)
        print('Free : {0}'.format(libName))
    else:
        print('No Library {0} Found'.format(libName))

def freeAllLibs(doAll=False):
    """
    libère toutes les librairies chargées dynamiquement
    """
    for libFile in bge.logic.LibList():
        libName = libFile.replace('.blend', '')
        libName = libName.split(SEP)[-1]
        if not(doAll):
            if not(libName in allwaysObjects['ALL_GAMES']):
                bge.logic.LibFree(libFile)
                print('Free : {0}'.format(libName))
            else:
                print('libName in allwaysObjects :', libName)
        else:
            bge.logic.LibFree(libFile)
            print('Free : {0}'.format(libName))

def loadChar(charName, mustReload=False):
    """
    chargement d'un personnage
    """
    root = bge.logic.expandPath('//')
    characterFile = '{1}..{0}chars{0}{2}.blend'.format(SEP, root, charName)
    characterFile = bge.logic.expandPath(characterFile)
    if not(characterFile in bge.logic.LibList()) or mustReload:
        try:
            bge.logic.LibLoad(characterFile, 'Scene', load_actions=True)
            print('Loaded : {0}'.format(charName))
        except:
            print('ERROR : {0}'.format(charName))
    else:
        print('Library {0} already loaded'.format(charName))



###########################################################'
#
#       AFFICHAGE DES TEXTES :
#       d'après :
#       http://www.highspeedracer.webou.net/tuto.php?id=33
#
###########################################################'

# les couleurs utilisées pour les textes :
colors = {
    'orange': (1.0, 0.5, 0.25, 1.0), 
    'blue': (0.0, 0.0, 1.0, 1.0), 
    'violet': (1.0, 0.0, 1.0, 1.0),
    'white': (1.0, 1.0, 1.0, 1.0),
    'gray': (0.5, 0.5, 0.5, 1.0),
    'red': (1.0, 0.0, 0.0, 1.0),
    }

def loadFonts():
    """
    charge les polices dans un dictionnaire
    elles doivent être dans le sous-dossier fonts
    """
    fontsNames = ('GoodDog.otf', 'Comfortaa_Bold.ttf')
    if not('FONTS' in bge.logic.globalDict):
        bge.logic.globalDict['FONTS'] = {0: 0}
        root = bge.logic.expandPath("//")
        i = 0
        for fontName in fontsNames:
            fontFile = '{1}..{0}fonts{0}{2}'.format(SEP, root, fontName)
            fontFile = bge.logic.expandPath(fontFile)
            i += 1
            bge.logic.globalDict['FONTS'][i] = blf.load(fontFile)

def print2D(text, pos, color=(0, 0, 0, 1), size=1, font=0):
    # récupération du numéro de la police :
    if not(font in bge.logic.globalDict['FONTS']):
        font = 0
    font = bge.logic.globalDict['FONTS'][font]
    # Recupere les informations de l'affichage - Collect viewport information
    view_buf = Buffer(GL_INT, 4)
    glGetIntegerv(GL_VIEWPORT, view_buf)
    view = view_buf

    # Parametre les matrices OpenGL - Setup OpenGL matrices
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    gluOrtho2D(0, view[2], 0, view[3])

    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()

    # On desactive les ombres - Disable shadows
    glDisable(GL_LIGHTING)

    height = bge.render.getWindowHeight()
    width = bge.render.getWindowWidth()

    # Taille du texte - Text size
    blf.size(font, int(height / 20), int(50 * size))
    glColor4f(color[0], color[1], color[2], color[3])
    # On écrit le texte en le découpant en lignes :
    lines = 0
    for line in text.split('|'):
        if lines > 0:
            decalage = 1.5 * int(lines * 1.5 * height / 20)
        else:
            decalage = 0
        blf.position(font, pos[0] * width, (1 - pos[1]) * height - decalage, 0)
        blf.draw(font, line)
        lines += 1

    # On reactive les ombres - Enable shadows
    glEnable(GL_LIGHTING)

def setText(textName='text1', text=None, time=None, color=None, width=None, position=None, font=None):
    if not('TEXTS' in bge.logic.globalDict):
        bge.logic.globalDict['TEXTS'] = {}
    if textName in bge.logic.globalDict['TEXTS']:
        # on ne modifie que ce qui est passé en paramètre :
        if text == None:
            text = bge.logic.globalDict['TEXTS'][textName][0]
        if time == None:
            time = bge.logic.globalDict['TEXTS'][textName][1]
        if color == None:
            color = bge.logic.globalDict['TEXTS'][textName][2]
        if width == None:
            width = bge.logic.globalDict['TEXTS'][textName][3]
        if position == None:
            position = bge.logic.globalDict['TEXTS'][textName][4]
        if font == None:
            font = bge.logic.globalDict['TEXTS'][textName][5]
    else:
        # on met de vraies valeurs pour les paramètres absents :
        if text == None:
            text = ''
        if time == None:
            time = -1
        if color == None:
            color = colors['white']
        if width == None:
            width = 1
        if position == None:
            position = (0.0, 0.5)
        if font == None:
            font = 1
    bge.logic.globalDict['TEXTS'][textName] = [text, time, color, width, position, font]

def updateTexts():
    finishedTexts =  []
    try:
        for textName in bge.logic.globalDict['TEXTS']:
            time = bge.logic.globalDict['TEXTS'][textName][1]
            if time > 0:
                time -= 1
                if time == 0:
                    bge.logic.globalDict['TEXTS'][textName][0] = ''
                    finishedTexts.append(textName)
                bge.logic.globalDict['TEXTS'][textName][1] = time
    except:
        if not('TEXTS' in bge.logic.globalDict):
            bge.logic.globalDict['TEXTS'] = {}
    finally:
        return finishedTexts

def postDrawTexts(state=''):
    """
    affichage des textes à l'écran.
    Procédure effectuée à la fin du rendu (scene.post_draw += [postDrawTexts]).
    """
    if state != '':
        if bge.logic.globalDict['GAME']['state'] != state:
            return
    # les textes affichés à l'écran :
    for textName in bge.logic.globalDict['TEXTS']:
        params = bge.logic.globalDict['TEXTS'][textName]
        print2D(params[0], params[4], params[2], params[3], params[5])

def createMenuItem(menuItem, name='', text='', comment='',
                width=3, position=(0, 0), what=-1, ok=True):
    """
    création d'un élément de menu
    """
    if not ('menuItems' in bge.logic.globalDict['TEMP']):
        bge.logic.globalDict['TEMP']['menuItems'] = {}
    bge.logic.globalDict['TEMP']['menuItems'][menuItem] = {
        'name': name,
        'text': text,
        'comment': comment,
        'width': width,
        'initialWidth': width,
        'position': position,
        'what': what,
        'ok': ok,
        }
    if ok:
        color = colors['orange']
    else:
        color = colors['gray']
    setText(menuItem, text=text, color=color, width=width, position=position)

def doMessage(message=''):
    """
    Permet d'écrire un message avec accents.
    On utilise un process différent et le message est enregistré dans
    le fichier message.txt.
    Enfin on lit le fichier pour récupérer le message.
    """
    installedPython = bge.logic.globalDict['OTHER']['installedPython']
    if installedPython == '':
        return ''
    import time
    import subprocess
    messageText = message.replace('\n', '|')
    args = []
    root = bge.logic.expandPath('//')
    lang = bge.logic.globalDict['MENU']['lang']

    processFile = '{0}..{1}scripts{1}messages.py'.format(root, SEP)
    processFile = bge.logic.expandPath(processFile)
    args = [
        installedPython, 
        processFile, 
        'LANG={0}'.format(lang), 
        'TEXT="{0}"'.format(messageText)]
    """
    if sys.platform.startswith('win'):
        # vu que sous windows Python n'est pas forcément installé,
        # j'ai dû faire un exe :
        processFile = '{0}..{1}scripts{1}winexe{1}messages.exe'.format(root, SEP)
        processFile = bge.logic.expandPath(processFile)
        args = [
            processFile, 
            'LANG={0}'.format(lang), 
            'TEXT="{0}"'.format(messageText)]
    else:
        # on lance le fichier messages.py :
        processFile = '{0}..{1}scripts{1}messages.py'.format(root, SEP)
        processFile = bge.logic.expandPath(processFile)
        args = [
            installedPython, 
            processFile, 
            'LANG={0}'.format(lang), 
            'TEXT="{0}"'.format(messageText)]
    """
    try:
        bge.logic.globalDict['GAME']['pause'] = True
        process = subprocess.Popen(args)
        while process.poll() == None:
            time.sleep(0.1)
        del process
        messageText = loadMessageText()
    except:
        print('DO_MESSAGE FAILED')
    if messageText != '':
        messageText = messageText.replace('\n', '|')
    bge.logic.globalDict['GAME']['pause'] = False
    return messageText

def loadMessageText():
    """
    lit le contenu du fichier message.txt
    """
    messageText = ''
    try:
        messageFileName = os.path.join(findConfigDir(), 'message.txt')
        theFile = open(messageFileName, newline='', encoding='utf-8')
        messageText = ''.join(theFile.readlines())
        theFile.close()
    except:
        pass
    return messageText



###########################################################'
#
#       TRADUCTIONS VIA DES FICHIERS CSV :
#
###########################################################'

def tr(csvFile, text, lang):
    """
    traduit le texte dans la langue demandée.
    Si la traduction n'est pas trouvée, on retourne le texte de départ.
    """
    result = text
    # on vérifie que le bon fichier est ouvert, sinon on l'ouvre :
    if not(csvFile in bge.logic.globalDict['TRANSLATIONS']['csvFiles']):
        openCsvFile(csvFile)
    # on lit la traduction :
    if text in bge.logic.globalDict['TRANSLATIONS']['tr']:
        result = bge.logic.globalDict['TRANSLATIONS']['tr'][text].get(lang, text)
    return result

def openCsvFile(csvFile):
    languages = []
    root = bge.logic.expandPath('//')
    csvFileName = '{1}..{0}translations{0}{2}.csv'.format(SEP, root, csvFile)
    csvFileName = bge.logic.expandPath(csvFileName)
    theFile = None
    try:
        bge.logic.globalDict['TRANSLATIONS']['csvFiles'].append(csvFile)
        nbCols = 0
        theFile = open(csvFileName, newline='', encoding='utf-8')
        reader = csv.reader(
            theFile, 
            delimiter=';', 
            quotechar='"', 
            quoting=csv.QUOTE_NONNUMERIC)
        firstRow = True
        for row in reader:
            if firstRow:
                firstRow = False
                for lang in row[1:]:
                    languages.append(lang)
                nbCols = len(languages)
            else:
                text = row[0]
                bge.logic.globalDict['TRANSLATIONS']['tr'][text] = {}
                for numCol in range(nbCols):
                    lang = languages[numCol]
                    translation = row[numCol + 1]
                    if translation == '':
                        bge.logic.globalDict['TRANSLATIONS']['tr'][text][lang] = text
                    else:
                        bge.logic.globalDict['TRANSLATIONS']['tr'][text][lang] = translation
        if csvFile == 'menu':
            bge.logic.globalDict['TRANSLATIONS']['languages'] = languages
    except:
        print('Failed to open {0}'.format(csvFileName))
    finally:
        if theFile != None:
            theFile.close()



###########################################################'
#
#       TÉLÉCHARGEMENTS DE FICHIERS :
#
###########################################################'

exhibitionWalls = ['salle_1', 'salle_2', 'salle_3', 'salle_4']

def downloadFilesExhibition():
    """
    mise à jour des fichiers textures du jeu exposition (Exhibition).
    """
    # on commence par remplir les dicos avec des dates bidons :
    oldDates = {}
    newDates = {}
    for wall in exhibitionWalls:
        oldDates[wall] = 10000101
        newDates[wall] = 10000101
    datesFileName = os.path.join(
        bge.logic.globalDict['downloadDir'], 'dates.txt')
    # si les fichiers n'existent pas encore, on récupère ceux fournis
    # dans le dossier /college3d/game/data/expo/salles :
    if not(os.path.isfile(datesFileName)):
        import shutil
        destDir = bge.logic.globalDict['downloadDir']
        root = bge.logic.expandPath('//')
        srcDir = '{1}..{0}data{0}expo{0}salles{0}'.format(SEP, root)
        shutil.copy('{0}dates.txt'.format(srcDir), destDir)
        for wall in exhibitionWalls:
            shutil.copy('{0}{1}.jpeg'.format(srcDir, wall), destDir)
    # on vérifie la date du dernier test 
    # pour ne pas mettre à jour plus d'une fois par jour :
    today = bge.logic.globalDict['SEASONS']['today']
    today = today[0]*10000 + today[1]*100 + today[2]
    lastTest = bge.logic.globalDict['OTHER']['lastDownload']
    if today - lastTest < 1:
        return
    bge.logic.globalDict['OTHER']['lastDownload'] = today
    # on récupère les dates actuelles :
    datesFile = open(datesFileName, encoding='utf-8')
    for a_line in datesFile:
        line = a_line.rstrip().split(':')
        if len(line) == 2:
            oldDates[line[0]] = int(line[1])
    datesFile.close()
    # on télécharge les nouvelles dates :
    url = '{0}/dates.txt'.format(DOWNLOAD_DIR)
    r = downloadFile(url)
    print('downloadFile :', url, r)
    datesFile = open(datesFileName, encoding='utf-8')
    for a_line in datesFile:
        line = a_line.rstrip().split(':')
        if len(line) == 2:
            newDates[line[0]] = int(line[1])
    datesFile.close()
    # on compare et télécharge les fichiers si besoin :
    files2Download = []
    for fileName in oldDates:
        if oldDates[fileName] < newDates[fileName]:
            files2Download.append(fileName)
    if len(files2Download) > 0:
        print('files2Download :', files2Download)
    for fileName in files2Download:
        url = '{0}/{1}.jpeg'.format(DOWNLOAD_DIR, fileName)
        downloadFile(url)

def downloadFile(url):
    """
    téléchargement d'un fichier.
    """
    from urllib import request
    filename = url.split('/')[-1]
    destFileName = os.path.join(
        bge.logic.globalDict['downloadDir'], filename)
    try:
        request.urlretrieve(url, destFileName)
        return True
    except:
        return False

def createTexture(obj, fileDir, textureName, fileName='', fileExt='jpeg'):
    """
    """
    if fileName == '':
        fileName = textureName
    # get the reference pointer (ID) of the internal texture
    matID = bge.texture.materialID(obj, 'IM{0}.{1}'.format(textureName, fileExt))
    # create a texture object
    object_texture = bge.texture.Texture(obj, matID)
    # create a new source with an external image
    newFileName = os.path.join(
        fileDir, '{0}.{1}'.format(fileName, fileExt))
    newFileName = bge.logic.expandPath(newFileName)
    new_source = bge.texture.ImageFFmpeg(newFileName)
    # the texture has to be stored in a permanent Python object
    bge.logic.globalDict['TEXTURES'][textureName] = object_texture
    # update/replace the texture
    bge.logic.globalDict['TEXTURES'][textureName].source = new_source
    bge.logic.globalDict['TEXTURES'][textureName].refresh(False)

def removeTexture(textureName):
    """
    Delete the Dynamic Texture, reversing back the final to its original state.
    """
    try:
        del bge.logic.globalDict['TEXTURES'][textureName]
    except:
        pass



###########################################################'
#
#       FONCTIONS DIVERSES :
#
###########################################################'

def u(text):
    """
    retourne une version unicode de text
    """
    try:
        if isinstance(text, str):
            return text
        else:
            return str(text)
    except:
        print('ERROR utils.u', type(text), text)
        return text

def openDir(dirName):
    """
    ouvre le dossier dirName avec le logiciel adéquat
    """
    if sys.platform.startswith('win'):
        app = 'explorer'
    elif sys.platform.startswith('darwin'):
        return
    else:
        app = 'xdg-open'
    try:
        commandLine = u('{0} "{1}"').format(app, dirName)
        os.system(commandLine)
    except:
        pass

def playSound(soundName='clic'):
    """
    joue un bruitage
    """
    if not(bge.logic.globalDict['MENU']['sounds']):
        return
    controller = bge.logic.getCurrentController()
    if soundName == 'clic':
        soundActuator = controller.actuators['Clic']
    elif soundName == 'over':
        soundActuator = controller.actuators['Over']
    else:
        soundActuator = None
    if soundActuator != None:
        controller.activate(soundActuator)

def makeScreenshot():
    """
    fait une photo d'écran du jeu (pour faire une vidéo)
    """
    n = bge.logic.globalDict['VIDEO']['CompteurScreenshot']
    n += 1
    if n < 3:
        bge.logic.globalDict['VIDEO']['CompteurScreenshot'] = n
        return
    bge.logic.globalDict['VIDEO']['CompteurScreenshot'] = 0
    n = bge.logic.globalDict['VIDEO']['NumScreenshot']
    if n < 10:
        num = '00000{0}'.format(n)
    elif n < 100:
        num = '0000{0}'.format(n)
    elif n < 1000:
        num = '000{0}'.format(n)
    elif n < 10000:
        num = '00{0}'.format(n)
    elif n < 100000:
        num = '0{0}'.format(n)
    else:
        num = n
    fileName = os.path.join(
        bge.logic.globalDict['VIDEO']['videoDir'], 
        'screenshot_{0}.png'.format(num))
    bge.render.makeScreenshot(fileName)
    bge.logic.globalDict['VIDEO']['NumScreenshot'] += 1

def nextInList(theList, actualItem):
    """
    retourne l'élément suivant dans la liste,
    ou le premier élémént si on est à la fin
    """
    if not(actualItem in theList):
        newItemIndex = 0
    else:
        newItemIndex = theList.index(actualItem) + 1
        if newItemIndex == len(theList):
            newItemIndex = 0
    return theList[newItemIndex]

def doRound(number, precision=3):
    """
    retourne la version arrondie d'un nombre selon la précision demandée
    """
    puissance = 10 ** precision
    return round(puissance * number) / puissance


